/* #define DEBUG 1 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>
#include <string.h>
#include <malloc.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>

#include <sys/types.h>
#include <sys/stat.h>

extern char **environ;

#define CONFIG_FILENAME "/etc/runsuid.rc"
#define EXEC_FILENAME	"/usr/sbin/runsuid"	/* only used on broken OSes */
#define EXEC_NAME	"runsuid"
#define MAX_ENV_NUM	63			/* maximum number of vars */


#ifdef DEBUG
#  define DBG(x)	do { fprintf x; } while (0)
#  define EXITERR(x)	do { fprintf (stderr, "%s:%d: ", __FILE__, __LINE__); fprintf x; if (errno != 0) perror ("Errno"); exit (2); } while (0)
#else
#  define DBG(x)	((void)0)
#  define EXITERR(x)	do { fprintf x; if (errno != 0) perror ("Errno"); exit (2); } while (0)
#endif


/* real user id / group id */
uid_t ruid;
gid_t rgid;
/* effective user id / group id */
uid_t euid;
gid_t egid;
/* all groups of real user */
gid_t *rgroups;
int   nrgroups;

/* environment: buffers */
char *execenv  [MAX_ENV_NUM+1] = { NULL };
char ruserenv [20];			/* REAL_USER= + 8 chars + \0 */
char euserenv [16];			/* USER= + 8 chars + \0 */

/* For error output: configuration file and current line number */
const char cfgname [] = CONFIG_FILENAME;
int  configline = 0;


/* Get uid by getpwnam / atoi */
uid_t getUserId (const char *string)
{
    uid_t r;
    struct passwd *p = getpwnam (string);
    if (p)
	return p->pw_uid;
    r = atoi (string);
    if (r == 0 && strcmp (string, "0") != 0)
	EXITERR ((stderr, "runsuid: Bad UserId '%s' in configuration file, line %d\n", string, configline));
    return r;
}


/* Get gid by getgrnam / atoi */
gid_t getGroupId (const char *string)
{
    gid_t r;
    struct group *g = getgrnam (string);
    if (g)
	return g->gr_gid;
    r = atoi (string);
    if (r == 0 && strcmp (string, "0") != 0)
	EXITERR ((stderr, "runsuid: Bad GroupId '%s' in configuration file, line %d\n", string, configline));
    return r;
}


/* Check real uid/gid for user:group -> 1: ok  0: failed
 * Side effects:
 * !!! Changes contents of <us> !!! */
int checkUserGroup (char *us)
{
    char *gs = strchr (us, ':');

    if (! gs)
	EXITERR ((stderr, "runsuid: Bad user:group definition '%s' in configuration file, line %d\n", us, configline));
    *gs++ = '\0';			/* Seperate user and group */

    /* fail, if user given and not identically */
    if (*us && getUserId (us) != ruid)
	return 0;

    /* fail, if group given and not identical or in user groups list */
    /* succeed otherwise */
    if (*gs)
    {
	int i;
	gid_t gc = getGroupId (gs);
	if (gc == rgid)
	    return 1;
	for (i = 0; i < nrgroups; i++)
	    if (gc == rgroups [i])
		return 1;
	return 0;
    }
    return 1;
}


/* Set euid / egid to user:group
 * Side effects:
 * !!! Changes contents of <us> !!! */
void setUserGroup (char *us)
{
    char *gs = strchr (us, ':');

    if (! gs)
	EXITERR ((stderr, "runsuid: Bad user:group definition '%s' in configuration file, line %d\n", us, configline));
    *gs++ = '\0';			/* Seperate user and group */
    euid = ruid;
    egid = rgid;

    /* get given ids and change invocation user/group ids */
    if (*us)
	euid = getUserId (us);
    if (*gs)
	egid = getGroupId (gs);
}


/* Check for wildcard matching -> 1: ok  0: failed */
int checkWildcard (const char *pattern, const char *string)
{
    const char *c, *cc, *s;

    /*
     * c: current pattern part (starting with '/') to be checked
     * cc: end of current pattern part to be checked
     * s: current string  part (starting with '/') to be checked
     */
    /* Perform this loop as long as at least one part (c and/or s) exists */
    /* Only if both parts exist, we are allowed to continue (strchr) -
     * see 1st if in loop */
    for (c = pattern, s = string; c || s;
	 c = strchr (c+1, '/'), s = strchr (s+1, '/'))
    {
	/* if one of these strings does not exist, one pattern is longer
	 * -> [failed] */
	if (!c || !s)
	    return 0;
	/* The string part that is now to be checked is either the length to
	 * the next '/' or the remaining string, whatever is less */
	/* cc points to the first character, that is *not* to be checked any
	 * more */
	cc = strchr (c+1, '/');
	if (cc == 0)
	    cc = pattern + strlen (pattern);
	/* case 1: c='/ * /...' etc. wildcard matches any part in s */
	if (c[1] == '*' && c+2 == cc)
	    (void) 0;	/* no, I'm not on drugs 8-P - it is cleaner this way */
	/* case 2: compare path components, that are no wildcards */
	/* 1st: strings must match. 2nd: s must continue with '/'or '\0' */
	else if (strncmp (c, s, cc-c) != 0 ||
		 (s [cc-c] != '/' && s [cc-c] != '\0'))
	    return 0;
    }
    /* If both strings are finished at the same time, the test has passed
     * -> [ok] */
    return 1;
}


/* Search calling environment for var */
char *searchEnv (const char *name)
{
    char **var = environ;
    int  len   = strlen (name);
    while (*var)
	if (strncmp (*var, name, len) == 0)
	    return *var;
	else
	    var++;
    return NULL;
}


/* Replace/insert/delete env var <name> in execenv by <line>.
 * Name has to be '<varname>='.
 * Content has to be NULL (delete) or '<varname>=<content>'.
 * Side effects:
 * !!! saves pointer to <line> in execenv !!! */
void replaceEnv (const char *name, char *line)
{
    char **var = execenv;
    int len    = strlen (name);
  
    DBG ((stderr, "runsuid: replaceEnv %s by %s\n",
	  name, line ? line : "NULL"));
    /* check whether var already exists */
    while (*var)
	if (strncmp (*var, name, len) == 0)
	{
	    if (line)
		*var = line;		/* Replace */
	    else
	    {
		do {			/* Delete */
		    *var = *(var+1);
		} while (*++var);
	    }
	    return;
	}
	else
	    var++;
    /* Insert */
    if (var - execenv >= MAX_ENV_NUM)
	EXITERR ((stderr, "runsuid: maximum environment size %d exceeded in configuration file %s, line %d\n", MAX_ENV_NUM, cfgname, configline));
    *var++ = line;
    *var   = NULL;
}


/* Set euid + egid and the environment according to config line.
 * Side effects:
 * !!! Changes contents of <usergroup> and <options> !!! */
void setIdEnv (char *usergroup, char *options)
{
    struct passwd *p;
    char *var, *content, *dup;

    /* Parse user:group string and set euid/egid */
    setUserGroup (usergroup);

    /* Inherrit some env variables */
    replaceEnv ("HOME=",    searchEnv ("HOME="));
    replaceEnv ("LOGNAME=", searchEnv ("LOGNAME="));
    replaceEnv ("TERM=",    searchEnv ("TERM="));

    /* Set some env vars to built in defaults */
    replaceEnv ("SHELL=", "SHELL=/bin/sh");
    replaceEnv ("PATH=", "PATH=/usr/sbin:/usr/bin:/usr/bsd:/sbin:/bin");
    replaceEnv ("LANG=", "LANG=C");

    /* set REAL_USER env */
    p = getpwuid (ruid);
    if (! p || ! p->pw_name)
	EXITERR ((stderr, "runsuid: Internal error on getting gid for real user\n"));
    sprintf (ruserenv, "REAL_USER=%.8s", p->pw_name);
    replaceEnv ("REAL_USER=", ruserenv);

    /* set USER env */
    p = getpwuid (euid);
    if (! p || ! p->pw_name)
	EXITERR ((stderr, "runsuid: Internal error on getting gid for effective user\n"));
    sprintf (euserenv, "USER=%.8s", p->pw_name);
    replaceEnv ("USER=", euserenv);
  
    /* set env vars according to options */
    while (options)
    {
	while (isspace (*options))
	    options++;
	if (! *options)
	    break;
	var = options;
	while (*options && ! isspace (*options) && *options != '=')
	    options++;
	if (*options == '=')
	{
	    options++;
	    /* memory layout when entering:
	     * ENVNAME="CONTENT_LINE"...  or ENVNAME=+... (or -)
	     * ^var    ^options              ^var    ^options */
	    switch (*options) {
	    case '"':
		content = options;
		*options++ = '\0';
		if (! (dup = strdup (var)) )
		    EXITERR ((stderr, "strdup failed"));
		while (*options && *options != '"')
		    options++;
		if (*options != '"')
		    EXITERR ((stderr, "runsuid: malformed option begining with '%s' in configuration file %s, line %d\n", var, cfgname, configline));
		*options++ = '\0';
		memmove (content, content+1, strlen (content+1) +1);
		/* memory layout now:
		 * ENVNAME=CONTENT_LINE\0\0...         ENVNAME=\0
		 * ^var    ^content        ^options    ^dup (via strdup) */
		/* This is not really neccessary (we could use the pointer <var> to
		 * save the value), but the simple approach would be vulnerable to
		 * programming errors and we would need a static read buffer */
		if (! (content = strdup (var)) )
		    EXITERR ((stderr, "strdup failed"));
		/* memory layout now:
		 * ENVNAME=LINE\0\0...        ENVNAME=\0    ENVNAME=LINE\0
		 *                 ^options   ^dup (strdup) ^content (strdup) */
		replaceEnv (dup, content);
		free (dup);
		break;
	    case '+':
		*options++ = '\0';
		replaceEnv (var, searchEnv (var));
		break;
	    case '-':
		*options++ = '\0';
		replaceEnv (var, NULL);
		break;
	    default:
		EXITERR ((stderr, "runsuid: malformed option begining with '%s' in configuration file %s, line %d\n", var, cfgname, configline));
	    }
	}
	else if (strncmp (var, "LOCALE", 6) == 0 && options == var+6)
	{
	    const char * const localevars[] = {
		"LANG=", "LC_CTYPE=", "LC_NUMERIC=", "LC_TIME=", "LC_COLLATE=",
		"LC_MONETARY=", "LC_MESSAGES=", "LC_PAPER=", "LC_NAME=",
		"LC_ADDRESS=", "LC_TELEPHONE=", "LC_MEASUREMENT=",
		"LC_IDENTIFICATION=", "LC_ALL=",
		NULL
	    };
	    const char * const *c;

	    for (c = localevars; *c; c++)
		replaceEnv (*c, searchEnv (*c));
	}
	else
	    EXITERR ((stderr, "runsuid: malformed option begining with '%s' in configuration file %s, line %d\n", var, cfgname, configline));
    }
}


/* check whether we are allowed to call the script and set euid + egid
 * and the environment accordingly */
void checkSetPermission (char *script)
{
    FILE *f;
    struct stat s;
    char readbuf [1024];

    /* First check permissions of config file */
    if (lstat (cfgname, &s) < 0)
	EXITERR ((stderr, "runsuid: Cannot stat config file %s\n", cfgname));
    if (s.st_uid != 0 || s.st_gid != 0)
	EXITERR ((stderr, "runsuid: Config file %s does not belong root:root\n", cfgname));
    if (! S_ISREG (s.st_mode))
	EXITERR ((stderr, "runsuid: Config file %s is not a regular file\n", cfgname));
    if (s.st_mode & (S_IWGRP | S_IWOTH))
	EXITERR ((stderr, "runsuid: Configuration file %s is writable for group or others\n", cfgname));

    /* Ok, read in config file and check info for each line */
    if (! (f = fopen (cfgname, "r")))
	EXITERR ((stderr, "runsuid: Cannot read config file %s\n", cfgname));
    configline = 0;
    while (fgets (readbuf, 1020, f))
    {
	char *path = readbuf, *check, *invoke, *options, *c;

	configline++;
	/* Eat white space at the beginning */
	while (isspace (*path))
	    path++;
	/* Check for empty lines and comments */
	if (*path == '\0' || *path == '#')
	    continue;
	/* We have now: Script path */
	check = path;
	while (! isspace (*check) && *check)
	    check++;
	if (! *check)
	    EXITERR ((stderr, "runsuid: Short line (1 arg) in configuration file %s, line %d\n", cfgname, configline));
	*check++ = '\0';
	while (isspace (*check))
	    check++;
	/* We have now: User/Group Id to be checked */
	invoke = check;
	while (! isspace (*invoke) && *invoke)
	    invoke++;
	if (! *invoke)
	    EXITERR ((stderr, "runsuid: Short line (1 arg) in configuration file %s, line %d\n", cfgname, configline));
	*invoke++ = '\0';
	while (isspace (*invoke))
	    invoke++;
	/* We have now: User/Group Id to be set */
	options = invoke;
	c       = invoke;
	/* Scan for end of line termination */
	while (*c && *c != '\r' && *c != '\n')
	    c++;
	if (! *c)
	    EXITERR ((stderr, "runsuid: Unterminated or long line in configuration file %s, line %d\n",
		      cfgname, configline));
	while (! isspace (*options) && *options)
	    options++;
	if (*options)
	    *options++ = '\0';
	while (options && isspace (*options))
	    options++;
	if (! *options)
	    options = NULL;
	DBG ((stderr, "runsuid: path=%s check=%s invoke=%s options=%s",
	      path, check, invoke, options ? options: "NULL\n"));
	/* Ok, analyzed string, now check for script path */
	if (! checkWildcard (path, script))
	    continue;
	/* Now check and set user/group id and environment */
	if (checkUserGroup (check))
	{
	    setIdEnv (invoke, options);
	    fclose (f);
	    return;
	}
    }
    /* Done, no match so far */
    if (ferror (f))
	EXITERR ((stderr, "runsuid: Error reading configuration file %s\n", cfgname));
    fclose (f);
    EXITERR ((stderr, "runsuid: User %s is not allowed to call script %s\n",
	      getpwuid (getuid ()) ->pw_name, script));
}


/* Initialize (static) data for real user */
void initUser (void)
{
    struct passwd *p = getpwuid (getuid ());
    if (! p)
	EXITERR ((stderr, "runsuid: Internal error on getting gid for real user\n"));
    ruid     = p->pw_uid;
    rgid     = p->pw_gid;
    nrgroups = getgroups (0, NULL);
    rgroups  = malloc (nrgroups * sizeof (gid_t));
    if (getgroups (nrgroups, rgroups) < nrgroups)
	EXITERR ((stderr, "runsuid: Internal error on reading group list for real user\n"));
#ifdef DEBUG
    {
        int i;
        fprintf (stderr, "runsuid: Called as uid %d gid %d groups", ruid, rgid);
        for (i = 0; i < nrgroups; i++)
	    fprintf (stderr, " %d", rgroups[i]);
	fprintf (stderr, "\n");
    }
#endif
}


/* main function */
int main (int argc, char *argv[])
{
    char buf [1024];
    FILE *f;
    char *s;
    char **execargs;

#ifdef DEBUG
    /* Debug output */
    {
	char **c;
	fprintf (stderr, "runsuid: Called as:\nrunsuid:   ");
	for (c = argv; *c; c++)
	    fprintf (stderr, " %s", *c);
	fprintf (stderr, "\n");
    }
#endif

    /* Usage */
    if (argc < 3)
	EXITERR ((stderr, "runsuid V" VERSIONSTR "\n"
		  "(c) 2000 Matthias Hopf <runsuid@mshopf.de>\n"
		  "Usage: %s <interpreter+args> <script> [<args> ...]\n"
		  "Check %s first, whether the script is allowed"
		  " to be run setuid/setgid.\n",
		  argv [0], cfgname));

    /* Check whether the operating system has stripped the s-Bit upon script
     * call. Re-exec in this case. The programm name is used as a flag in
     * this case (Ugh...) */
    if (geteuid () != 0)
    {
	DBG ((stderr, "runsuid: Not yet root, re-execing.\n"));
	if (argv [0][0] != '-')
	{
	    s = argv [0];
	    argv [0] = "-" EXEC_NAME;
	    /* Some operating systems call us with argv[0] being the
	     * script (!) name, and not the interpreter. Use compile-time
	     * default in this case. */
	    if (! strstr (s, EXEC_NAME))
	    {
		DBG ((stderr, "runsuid: OS called us with weired argv[0].\n"
		      "runsuid: Using compile-time executable path "
		      EXEC_FILENAME ".\n"));
		s = EXEC_FILENAME;
	    }
	    /* We are not root yet, so no check required */
	    execv (s, argv);
	    EXITERR ((stderr, "runsuid: Cannot re-exec %s.\n", s));
	}
	EXITERR ((stderr, "runsuid: " EXEC_FILENAME " is not set-uid root.\n"));
    }
    /* Init */
    initUser ();
    if (access (argv [2], R_OK | X_OK) < 0)
	EXITERR ((stderr, "runsuid: Cannot execute %s.\n", argv [2]));
    /* Read in interpreter line of script */
    if (! (f = fopen (argv [2], "r")) )
	EXITERR ((stderr, "runsuid: Cannot open %s.\n", argv [2]));
    if (! fgets (buf, 1020, f))
	EXITERR ((stderr, "runsuid: Cannot read %s.\n", argv [2]));
    fclose (f);
    /* Check for being a script */
    if (buf [0] != '#' || buf [1] != '!')
	EXITERR ((stderr, "runsuid: %s is not a script.\n", argv [2]));
    if ( (s = strchr (buf, '\n')) )
	*s = '\0';
    if ( (s = strchr (buf, '\r')) )
	*s = '\0';
    /* Get interpreter and compare with command argument */
    s = buf + 2;
    while (isspace (*s))
	s++;
    while (! isspace (*s) && *s)
	s++;
    while (isspace (*s))
	s++;
    if (strcmp (argv [1], s) != 0)
	EXITERR ((stderr, "runsuid: Interpreter mismatch in %s.\n", argv [2]));
    /* Check permission according to config file and set environment */
    checkSetPermission (argv [2]);
    /* split interpreter + args into interpreter and args, if any are given */
    execargs = argv+1;
    if ( (s = strchr (argv[1], ' ')) )
    {
	*s++ = '\0';
	execargs = argv;
	/* we don't need argv [0] here any more */
	argv [0] = argv [1];
	argv [1] = s;
    }
    /* Now set uid / gid accordingly */
    if (setgid (egid) < 0)
	EXITERR ((stderr, "runsuid: Cannot set group id.\n"));
    if (setuid (euid) < 0)	/* no way back now... */
	EXITERR ((stderr, "runsuid: Cannot set user id.\n"));
#ifdef DEBUG
    /* Debug output */
    {
	char **c;
	fprintf (stderr, "runsuid: Calling as %d:%d :\nrunsuid:   ",
		 euid, egid);
	for (c = execargs; *c; c++)
	    fprintf (stderr, " %s", *c);
	fprintf (stderr, "\n");
	fprintf (stderr, "runsuid: Environment:\n");
	for (c = execenv; *c; c++)
	    fprintf (stderr, "  %s\n", *c);
    }
#endif
    execve (*execargs, execargs, execenv);
    EXITERR ((stderr, "runsuid: Cannot exec %s.\n", *execargs));
    return 1;
}
